; File......: CHGATTR.ASM
; Author....: Jeff Cheney
; Date......: January 10, 1992
; Revision..: 1.0
; 
; This is an original work by Jeff Cheney and is placed in the
; public domain.
;
; Modification history:
; ---------------------
;    First release.
;
; Purpose:
; -------
;    Change attribute bytes in SAVESCREENed strings.
;
; Syntax:
; ------
;    CHGATTR( <cOldStr>,<nStdClr>,[<nGetClr>],[<lSetBlink>] ) -> <cNewStr>
;
; Arguments:
; ---------
;    <cOldStr>   is the string saved with SAVESCREEN().
;    <nStdClr>   is the color number to use for standard colors.
;    <nGetClr>   is the color number to use for enhanced colors.
;    <lSetBlink> is the current setting of SETBLINK().
;
; Returns:
; -------
;    <cNewStr>   is the new string with the colors replaced. It can simply be
;                RESTSCREENed back into its old location. <cNewStr> will be
;                NIL if:
;                         - less than two parameters are passed
;                         - the first parameter is not passed by reference
;                         - dereferencing fails (low memory?)
;                         - the second parameter is not a number
;
; Description:
; -----------
;    This function is very good for changing the color of a window without
;    using direct screen writing, thereby respecting DISPBEGIN() and
;    DISPEND(). It can also be used to draw shadows around boxes. ChgAttr()
;    takes a string returned by the SAVESCREEN() function and replaces every
;    second byte (the video attribute) with the color requested. It senses
;    GET fields by replacing any color byte in the existing string that does
;    not match the FIRST color byte in the existing string with the <nGetClr>.
;    If <nGetClr> is not provided, ChgAttr() reverses <nStdClr>, respecting
;    the <lSetBlink> setting if it is provided. If <lSetBlink> is not provided,
;    bold backgrounds are disabled by default. If <nGetClr> is passed, the
;    setting of <lSetBlink> is ignored and <nGetClr> is respected. This 
;    function
;    is small because the string is modified entirely in Clipper's own data 
;    area
;    (the string is dereferenced in NanSpeak).
;
; Examples:
; -------
;    // This function takes the four window coordinates and the new window
;    // color as parameters and returns the previous window which can simply
;    // be RESTSCREENed when it becomes active again.
;    FUNCTION idleWin( lnStartRow,lnStartCol,lnEndRow,lnEndCol,lcNewColor )
;    LOCAL lcOldStr,lcNewStr
;    lcOldStr := SAVESCREEN( lnStartRow,lnStartCol,lnEndRow,lnEndCol )
;    lcNewStr := chgAttr( @lcOldStr,                                           
;
;                         clr2num( lcNewColor ),                               
;
;                         clr2num( SUBSTR( lcNewColor,                         
;
;                                          AT( ",",lcNewColor ) + 1 ) ),       
;
;                         SETBLINK()                                      )
;    RESTSCREEN( lnStartRow,lnStartCol,lnEndRow,lnEndCol,lcNewStr )
;    RETURN lcOldStr
;
;    // This function takes the raw window coordinates and a shadow colour as
;    // parameters and draws a shadow below and to the right of the window.
;    FUNCTION shadow( lnTop,lnLeft,lnBottom,lnRight,lcShadowClr )
;    LOCAL lcOldStr := SAVESCREEN( lnBottom + 1,lnLeft  + 1, ;
;                                  lnBottom + 1,lnRight + 1 )
;    LOCAL lcNewStr := chgAttr( @lcOldStr,clrConv( lcShadowClr ) )
;    DISPBEGIN()
;      RESTSCREEN( lnBottom + 1,lnLeft  + 1,lnBottom + 1,lnRight + 1,lcNewStr )
;      lcOldStr := SAVESCREEN( lnTop + 1,lnRight + 1,lnBottom,lnRight + 1 )
;      lcNewStr := chgAttr( @lcOldStr,clr2num( lcShadowClr ) )
;      RESTSCREEN( lnTop + 1,lnRight + 1,lnBottom,lnRight + 1,lcNewStr )
;    DISPEND()
;    RETURN NIL
;
; Requires:
; --------
;    CLR2NUM() (follows this function)
;

public       chgattr

extrn        __parInfo:far
extrn        __parc:far
extrn        __parcLen:far
extrn        __parL:far
extrn        __parNI:far
extrn        __ret:far
extrn        __retCLen:far
extrn        __storCLen:far

dgroup       group   _my_dataseg

_my_dataseg  segment public  'DATA'

stdColor     DB      0
getColor     DB      0
strLngth     DW      0
errorNo      DB      0
strSegment   DW      0
strOffset    DW      0
pCount       DB      0

_my_dataseg  ends

_my_codeseg  segment word public 'CODE'

             assume  cs:_my_codeseg, ds:dgroup, es:dgroup

chgattr      proc    far

saveRegs:    push    bp                  ; preserve registers for Clipper
             mov     bp,sp               ;   "       "       "       "
             push    ds                  ;   "       "       "       "
             push    es                  ;   "       "       "       "
             push    si                  ;   "       "       "       "
             push    di                  ;   "       "       "       "
             mov     errorNo,0           ; make sure this defaults to 0
getInfo:     mov     ax,0                ; number of parameters
             push    ax                  ; push to Clipper
             call    __parInfo           ; get parameter count
             add     sp,2                ; restore stack
             mov     pCount,al           ; save parameter count
             cmp     ax,2                ; is string at least 2
             jae     chkString           ; yes, continue
             mov     errorNo,1           ; error - less than 2 parameters
             jmp     restRegs            ; return NIL
chkString:   mov     ax,1                ; first parameter
             push    ax                  ; push to Clipper
             call    __parInfo           ; get first parameter data type
             add     sp,2                ; restore stack
             cmp     ax,33               ; is it a string passed by reference?
             je      getFirst            ; yes, continue
             mov     errorNo,2           ; error - not a string passed by
                                         ;   reference
             jmp     restRegs            ; return NIL
getFirst:    mov     ax,1                ; first parameter
             push    ax                  ; push to Clipper
             call    __parc              ; get address of character string
             add     sp,2                ; restore stack
             mov     strSegment,dx       ; save string's segment
             mov     strOffset,ax        ; save string's offset
getLenP1:    mov     ax,1                ; first parameter
             push    ax                  ; push to Clipper
             call    __parcLen           ; get length of first parameter
             add     sp,2                ; restore stack
             mov     strLngth,ax         ; store length of string to modify
             mov     bx,strOffset        ; place string offset into BX
             mov     cx,ax               ; copy length to counter register
deRefP1:     mov     ax,1                ; first parameter
             push    ax                  ; push to Clipper
             push    cx                  ; push string length to Clipper
             push    dx                  ; push segment of string to Clipper
             push    bx                  ; push offset  of string to Clipper
             call    __storCLen          ; dereference string
             add     sp,6                ; restore stack
             pop     ax                  ; get return value
                                         ;   (success = 1,fail = 0)
             cmp     ax,1                ; was dereferencing successful?
             je      getLen2             ; yes, get second parameter
             mov     errorNo,3           ; dereferencing failed
             jmp     restRegs            ; return NIL
getLen2:     mov     ax,2                ; query second parameter
             push    ax                  ; push to Clipper
             call    __parInfo           ; get type of second parameter
             add     sp,2                ; restore stack
             cmp     ax,2                ; is second parameter a numeric?
             je      getSecond           ; yes, continue
             mov     errorNo,4           ; second parameter not a number
             jmp     restRegs            ; pass back NIL
getSecond:   mov     ax,2                ; query second parameter
             push    ax                  ; push to Clipper
             call    __parNI             ; get address of second parameter
             add     sp,2                ; restore stack
             mov     stdColor,al         ; save colour byte
chkIfThird:  cmp     pCount,3            ; check if third parameter passed
             jb      revSecond           ; no, reverse second parameter
             jmp     chkThird            ; yes, retrieve third parameter
revSecond:   mov     ah,al               ; save a copy of stdColor
             and     al,7                ; isolate foreground bits
             mov     cl,4                ; multiply by 16
             rol     al,cl               ; transfer to background
             and     ah,10001000b        ; save blink/bold bits
             or      al,ah               ; set blink/bold bits in AL
             mov     ah,stdColor         ; get fresh copy of stdColor
             and     ah,01110000b        ; isolate background bits
             mov     cl,4                ; divide by 16
             ror     ah,cl               ;
             or      al,ah               ; set new foreground bits
             mov     getColor,al         ; save getColor
chkFourth:   cmp     pCount,4            ; were there 4 parameters passed?
             je      getType4            ; yes, get fourth parameter's type
             jmp     update              ; no, bypass this routine
getType4:    mov     ax,4                ; query fourth parameter
             push    ax                  ; pass to Clipper
             call    __parInfo           ; get fourth paramater's data type
             add     sp,2                ; restore stack
             cmp     ax,4                ; is it a logical?
             je      getFourth           ; yes, continue
             jmp     update              ; no, ignore fourth parameter
getFourth:   mov     ax,4                ; query fourth parameter
             push    ax                  ; push to Clipper
             call    __parL              ; get fourth parameter's value
             add     sp,2                ; restore stack
             cmp     ax,0                ; is it .F.
             je      revBlink            ; yes, exchange blink and bold bits
             jmp     update              ; no, ignore fourth parameter
revBlink:    mov     al,stdColor         ; get standard colour
             mov     ah,al               ; store a second copy
             and     al,00001111b        ; zero upper 4 bits in AL
             mov     cl,4                ; move lower 4 bits to upper 4 bits
             rol     al,cl               ; (change background to foreground)
             and     ah,11110000b        ; zero lower 4 bits in AH
             mov     cl,4                ; move upper 4 bits to lower 4 bits
             ror     ah,cl               ; (change foreground to background)
             or      al,ah               ; combine foreground and background
             mov     getColor,al         ; save new colour
             jmp     update              ; modify string
chkThird:    mov     ax,3                ; query third parameter
             push    ax                  ; push to Clipper
             call    __parInfo           ; get third parameter's type
             add     sp,2                ; restore stack
             cmp     ax,2                ; is it a numeric?
             je      getThird            ; yes, get value
             mov     al,stdColor         ; no, use stdColor and
             jmp     revSecond           ;   reverse foreground/background
getThird:    mov     ax,3                ; third parameter
             push    ax                  ; push to Clipper
             call    __parNI             ; get third parameter
             add     sp,2                ; restore stack
             mov     getColor,al         ; save colour byte
update:      mov     cx,strLngth         ; set CX to length of string
             shr     cx,1                ; divide by 2 (use every second byte)
             mov     es,strSegment       ; get string's segment
             mov     di,strOffset        ; get string's offset
             inc     di                  ; start with second byte in string
             mov     al,stdColor         ; load stdColor into register
                                         ;   (for speed)
             mov     ah,getColor         ; load getColor into register
                                         ;   (for speed)
             mov     bl,es:[di]          ; load assumed stdColor into a
                                         ;   register (for speed)
change:      cmp     es:[di],bl          ; is this position currently stdColor
             je      isStdColor          ; yes, jump to isStdColor
             mov     byte ptr es:[di],ah ; move new get color into attribute
                                         ;   position
             jmp     short next          ; move to next attribute byte
isStdColor:  mov     byte ptr es:[di],al ; move new stdColor into attribute
                                         ;   position
next:        inc     di                  ; move to next attribute byte
             inc     di                  ;
             loop    change              ; decrement CX and LOOP
restRegs:    pop     di                  ; restore registers used
             pop     si                  ;   "       "        "
             pop     es                  ;   "       "        "
             pop     ds                  ;   "       "        "
             pop     bp                  ;   "       "        "
             cmp     errorNo,0           ; was there a fatal error?
             jz      passBack            ; no, pass back new string
             call    __ret               ; yes, return NIL
             jmp     theEnd
passBack:    mov     cx,strLngth         ; load length of return string
             mov     dx,strSegment       ; load segment of this string
             mov     ax,strOffset        ; load offset of this string
             push    cx                  ; push to Clipper
             push    dx                  ;   "         "
             push    ax                  ;   "         "
             call    __retCLen           ; return string pointed to by DX:AX
                                         ;   that is CX bytes long
             add     sp,6                ; restore stack
theEnd:      ret

chgAttr      endp

_my_codeseg  ends
             end
