; File......: CLR2NUM.ASM
; Author....: Jeff Cheney
; Date......: January 10, 1992
; Revision..: 1.0
; 
; This is an original work by Jeff Cheney and is placed in the
; public domain.
;
; Modification history:
; ---------------------
;    First Release
;
; Purpose:
; -------
;    Converts a Clipper color string to a number.
;
; Syntax:
; ------
;    CLR2NUM( <cColorString>,[<lSetBlink>] ) -> <nColorNum>
;
; Arguments:
; ---------
;    <cColorString> is a a color string containing one or more color
;                   attributes.
;    <lSetBlink>    is the current setting of SETBLINK().
;
; Returns:
; -------
;    <nColorNum>    is the numeric representation of the first color
;                   attribute in the <cColorString>.
;
; Description:
; -----------
;    If <nColorNum> is greater than 255 an error has occured as follows:
;              257 = incorrect number of parameters
;              258 = second parameter not a logical value
;              259 = first parameter was empty string
;              260 = colour string exceeds 128 characters
;              261 = invalid attribute parameter
;
;    Parsing of <cColorString> ends at the first comma encountered.
;
; Examples:
; --------
;    nColorNo := clr2num( "W/N" )         // returns 7
;

public       clr2num

extrn        __parInfo:far
extrn        __parcLen:far
extrn        __parc:far
extrn        __parl:far
extrn        __retni:far

dgroup       group   _my_dataseg

_my_dataseg  segment public  'DATA'

fgBitMask    db               00h       ; place to store foreground while
                                        ;   we set the background's bits
colorStr     db      128 dup( 00h )     ; color string parameter
errorNo      db               00h       ; number of error encountered - 256
                                        ;   (see Header)
booleans     db               00h       ; bit 1 = blank requested
                                        ; bit 2 = underline requested
                                        ; bit 3 = inverse requested
                                        ; bit 4 = blink/background bold
                                        ;         requested
                                        ; bit 5 = is current bitMask
                                        ;         the background?
                                        ; bit 6 = bold foreground requested
counter      db               00h       ; length of colour string passed
colorNo      db               00h       ; return color number
blinkBold    db               01h       ; flag used to toggle the meaning
                                        ;   of the blink/background bold bit
                                        ;   (default is blink)
_my_dataseg  ends

_my_codeseg  segment word public 'CODE'

             assume  cs:_my_codeseg, ds:dgroup, es:dgroup

clr2num      proc    far

saveRegs:    push    bp                  ; save registers used
             mov     bp,sp               ;   "             "
             push    ds                  ;   "             "
             push    es                  ;   "             "
             push    si                  ;   "             "
             push    di                  ;   "             "
clearData:   xor     al,al               ; set al to 0
             mov     bx,seg _my_dataseg  ; point ES to data segment
             mov     es,bx               ;

             mov     bx,offset fgBitMask ; point BX to offset of bit mask 2
             mov     es:[bx],al          ; zero alternate bitMask
             mov     bx,offset errorNo   ; point BX to offset of error number
             mov     es:[bx],al          ; zero error code
             mov     bx,offset booleans  ; point BX to offset of logicals
             mov     es:[bx],al          ; zero booleans
getInfo:     mov     ax,0                ; number of parameters
             push    ax                  ; push to CLIPPER
             call    __parInfo           ; get number of parameters
             add     sp,2                ; restore stack
             cmp     ax,0                ; were there any parameters?
             je      badCount            ; exit with error if 0
             cmp     ax,2                ; compare number of parameters to 2
             ja      badCount            ; exit with error if > 2
             je      getFlag             ; set SETBLINK() flag if 2 params
             mov     blinkBold,01h       ; default to .T.
             jmp     getLength           ; continue
badCount:    mov     errorNo,01h         ; invalid number of parameters
             jmp     exit                ; restore registers and return
getFlag:     mov     ax,2                ; use second parameter
             push    ax                  ; push to Clipper
             call    __parInfo           ; get type of variable
             add     sp,2                ; restore stack
             cmp     ax,4                ; is it a logical value?
             jz      getSetting          ; yes, save current setting
badType:     mov     errorNo,02h         ; not a logical value
             jmp     exit                ; restore registers and return
getSetting:  mov     ax,2                ; get second parameter
             push    ax                  ; push to Clipper
             call    __parl              ; get setBlink() setting
             add     sp,2                ; restore stack
             mov     blinkBold,al        ; save setting
getLength:   mov     ax,1                ; first parameter
             push    ax                  ; push to CLIPPER
             call    __parcLen           ; get length of first parameter
             add     sp,2                ; restore stack
             cmp     ax,0                ; was string empty
             jne     saveLngth           ; no, save string length
badString:   mov     errorNo,03h         ; yes, exit with error 3
             jmp     exit                ; restore registers & return
saveLngth:   cmp     ax,128              ; see if length > 128
             ja      tooMany             ; yes, exit with error code
             mov     counter,al          ; save length
             jmp     getString           ; get first parameter
tooMany:     mov     errorNo,04h         ; length exceeds maximum of 128
             jmp     exit                ; restore registers and return
getString:   mov     ax,1                ; get first parameter
             push    ax                  ; push to CLIPPER
             call    __parc              ; get address of character string
             add     sp,2                ; restore stack - DX:AX points to string
             xor     ch,ch               ; set CH register to 0
             mov     cl,counter          ; get number of bytes to move
             push    ds                  ; save DS
             mov     ds,dx               ; source segment is DX
             mov     si,ax               ; source offset is AX
             mov     di,offset ds:colorStr
                                         ; point to colorStr variable in DS
             mov     bx,di               ; save address for getAttrib
             cld                         ; set direction foreward
             rep     movsb               ; move COUNTER bytes DS:SI to ES:DI
             pop     ds                  ; restore DS
getAttrib:   mov     cl,counter          ; length of color string
             xor     dx,dx               ; clear bit mask & booleans
             jmp     getByte             ; do not increment first loop
getParam:    inc     bx                  ; next byte in string
getByte:     mov     al,es:[ bx ]        ; get parameter character
             call    upperCase           ; convert AL to uppercase
chkBlack:    cmp     al,"N"              ; is it BLACK?
             jnz     chkBlue             ; no, continue checking
             and     dl,11111000b        ; zero bits 0 to 2
             jmp     getNext             ; save changed bit mask
chkBlue:     cmp     al,"B"              ; is it BLUE?
             jnz     chkGreen            ; no, continue checking
             or      dl,00000001b        ; set bit 0
             jmp     getNext             ; save changed bit mask
chkGreen:    cmp     al,"G"              ; is it GREEN?
             jnz     chkRed              ; no, continue checking
             or      dl,00000010b        ; set bit 1
             jmp     getNext             ; save changed bit mask
chkRed:      cmp     al,"R"              ; is it RED?
             jnz     chkWhite            ; no, continue checking
             or      dl,00000100b        ; set bit 2
             jmp     getNext             ; save changed bit mask
chkWhite:    cmp     al,"W"              ; is it WHITE?
             jnz     chkBlank            ; no, continue checking
             or      dl,00000111b        ; set bits 0 to 2
             jmp     getNext             ; save changed bit mask
chkBlank:    cmp     al,"X"              ; is it BLANK
             jnz     chkUnder            ; no, continue checking
             or      dh,00000001b        ; set BLANK = .T.
             jmp     getNext             ; save changed boolean byte
chkUnder:    cmp     al,"U"              ; is it UNDERLINE
             jnz     chkInverse          ; no, continue checking
             or      dh,00000010b        ; set UNDERLINE = .T.
             jmp     getNext             ; save changed boolean byte
chkInverse:  cmp     al,"I"              ; is it INVERSE video
             jnz     chkBlink            ; no, continue checking
             or      dh,00000100b        ; set INVERSE = .T.
             jmp     getNext             ; save changed boolean byte
chkBlink:    cmp     al,"*"              ; is it BLINK
             jnz     chkBold             ; no, continue checking
             test    dh,00010000b        ; are we setting foreground?
             jz      chkBlinkOn          ; yes, check SETBLINK() setting
             jmp     getNext             ; no, ignore this character
chkBlinkOn:  cmp     blinkBold,01h       ; is blinking enabled?
             jz      blinkOk             ; yes, set blink on
             jmp     getNext             ; no, ignore this character
blinkOk:     or      dh,00001000b        ; set BLINK = .T.
             jmp     getNext             ; save changed boolean byte
chkBold:     cmp     al,"+"              ; is it BOLD
             jnz     chkSlash            ; no, continue checking
             test    dh,00010000b        ; are we setting foreground?
             jz      isFore              ; yes, set foreground bold
             cmp     blinkBold,00h       ; no, is background bold allowed?
             jnz     notBold             ; no, ignore
             or      dh,00001000b        ; yes, set BACKGROUND BOLD ON
notBold:     jmp     getNext             ; save changed boolean byte
isFore:      or      dh,00100000b        ; set BOLD = .T.
             jmp     getNext             ; save changed boolean byte
chkSlash:    cmp     al,"/"              ; is it end of foreground?
             jnz     chkComma            ; no, continue checking
             test    dh,00010000b        ; are we setting foreground?
             jz      save_fg             ; yes, save byte
             jmp     badParam            ; no, exit with error
save_fg:     mov     fgBitMask,dl        ; save foreground mask
             xor     dl,dl               ; clear bit mask register
             or      dh,00010000b        ; set background flag to .T.
             jmp     getNext             ; save changed boolean byte
chkComma:    cmp     al,","              ; is it a comma
             jnz     chkSpace            ; no, continue checking
             mov     cl,1                ; reduce counter to end routine
             jmp     getNext             ; end parsing
chkSpace:    cmp     al," "              ; is it space?
             jz      getNext             ; yes, skip it
badParam:    mov     errorNo,05h         ; invalid attribute parameter
             jmp     exit                ; restore registers and return
getNext:     dec     cx                  ; decrease counter
             jcxz    buildByte           ; exit if counter = 0
             jmp     getParam            ; loop to get next parameter

buildByte:   test    dh,00010000b        ; were we working on the background?
             jz      setBlank            ; no, leave byte alone
             mov     cl,4                ; rotate bits to foreground position
             rol     dl,cl               ;
             or      dl,fgBitMask        ; add background bits
setBlank:    test    dh,00000001b        ; was BLANK requested?
             jz      setUnderLn          ; no, check for underline
             mov     al,dl               ; copy video attribute byte into al
             cmp     blinkBold,01h       ; is blinking enabled?
             jz      blinkOn             ; yes, make sure bit is not set
             and     al,11110000b        ; include BOLD bit
             jmp     back2fore           ; copy background colour to foreground
blinkOn:     and     dl,01111111b        ; set blink bit off
             and     al,01110000b        ; isolate background colour
back2fore:   mov     cl,4                ; move background to foreground
             shr     al,cl               ;
             and     dl,11111000b        ; wipe out foreground colour
             or      dl,al               ; apply bit mask to video byte
             jmp     saveByte            ; bypass other checks that may
                                         ;   affect visibility
setUnderLn:  test    dh,00000010b        ; was UNDERLINE requested?
             jz      setInverse          ; no, check if inverse requested
             and     dl,10001000b        ; wipe out background & foreground
             or      dl,00000001b        ; set blue foreground
setInverse:  test    dh,00000100b        ; was inverse requested
             jz      setBlink            ; no, check for blink
             mov     cl,4                ; yes, rotate attribute byte
             rol     dl,cl               ;
setBlink:    test    dh,00001000b        ; was blink requested?
             jz      setBold             ; no, check for bold
             or      dl,10000000b        ; set bit 7 on
setBold:     test    dh,00100000b        ; was bold requested?
             jz      saveByte            ; no, update video memory
             or      dl,00001000b        ; set bit 3 on

saveByte:    mov     colorNo,dl          ; save byte
             jmp     exit                ; restore registers and return

upperCase:   cmp     al,061h             ; is it <"a"
             jb      isUpper             ; yes, do not alter
             cmp     al,07ah             ; is it >"z"
             ja      isUpper             ; yes, do not alter
             sub     al,020h             ; lowercase, convert to uppercase
isUpper:     retn                        ; return to caller

exit:        xor     ax,ax              ; set AX to 0
             mov     al,errorNo         ; move errorNo into AL
             cmp     al,0h              ; check if there was an error
             jnz     setHigh            ; yes, add 256 to error code
             mov     al,colorNo         ; no, return colour number
             jmp     restRegs           ; restore registers

setHigh:     mov     ah,01h             ; add 256 to error code

restRegs:    pop     di                 ; restore registers used
             pop     si                 ;    "              "
             pop     es                 ;    "              "
             pop     ds                 ;    "              "
             pop     bp                 ;    "              "

passBack:    push    ax                 ; push return value
             call    __retni            ; return a numeric integer
             add     sp,2               ; restore stack
             ret                        ; end

clr2num      endp

_my_codeseg  ends
             end

; --- end of clr2num() -------------------------------------------------
